/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** BusVoodoo RS-232 mode (code)
 *  @file busvoodoo_rs232.c
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2018
 *  @note peripherals used: USART @ref busvoodoo_rs232
 */
/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdlib.h> // standard utilities
#include <string.h> // string utilities

/* STM32 (including CM3) libraries */
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/usart.h> // USART utilities

/* own libraries */
#include "global.h" // board definitions
#include "print.h" // printing utilities
#include "menu.h" // menu definitions
#include "busvoodoo_global.h" // BusVoodoo definitions
#include "busvoodoo_oled.h" // OLED utilities
#include "busvoodoo_rs232.h" // own definitions

/** @defgroup busvoodoo_rs232_usart USART peripheral used for RS-232 communication, using an RS-232 transceiver
 *  @{
 */
#define BUSVOODOO_RS232_USART 2 /**< USART peripheral */
/** @} */

/** mode setup stage */
static enum busvoodoo_rs232_setting_t {
	BUSVOODOO_RS232_SETTING_NONE,
	BUSVOODOO_RS232_SETTING_BAUDRATE,
	BUSVOODOO_RS232_SETTING_DATABITS,
	BUSVOODOO_RS232_SETTING_PARITY,
	BUSVOODOO_RS232_SETTING_STOPBITS,
	BUSVOODOO_RS232_SETTING_HWFLOWCTL,
	BUSVOODOO_RS232_SETTING_DONE,
} busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_NONE; /**< current mode setup stage */
/** RS-232 baud rate (in bps) */
static uint32_t busvoodoo_rs232_baudrate = 115200;
/** RS-232 data bits */
static uint8_t busvoodoo_rs232_databits = 8;
/** RS-232 parity setting */
static uint32_t busvoodoo_rs232_parity = USART_PARITY_NONE;
/** RS-232 stop bits setting */
static uint32_t busvoodoo_rs232_stopbits = USART_STOPBITS_1;
/** RS-232 hardware flow control setting (true = with hardware flow control, false = without hardware flow control */
static bool busvoodoo_rs232_hwflowctl = false;

/** setup RS-232 mode
 *  @param[out] prefix terminal prompt prefix
 *  @param[in] line terminal prompt line to configure mode
 *  @return if setup is complete
 */
static bool busvoodoo_rs232_setup(char** prefix, const char* line)
{
	bool complete = false; // is the setup complete
	if (NULL==line) { // first call
		busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_NONE; // re-start configuration
	}
	switch (busvoodoo_rs232_setting) {
		case BUSVOODOO_RS232_SETTING_NONE:
			snprintf(busvoodoo_global_string, LENGTH(busvoodoo_global_string), "baud rate in bps (1-2000000) [%u]", busvoodoo_rs232_baudrate);
			*prefix = busvoodoo_global_string; // ask for baud rate
			busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_BAUDRATE;
			break;
		case BUSVOODOO_RS232_SETTING_BAUDRATE:
			if (NULL==line || 0==strlen(line)) { // use default setting
				busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_DATABITS; // go to next setting
			} else { // setting provided
				uint32_t baudrate = atoi(line); // parse setting
				if (baudrate>0 && baudrate<=2000000) { // check setting
					busvoodoo_rs232_baudrate = baudrate; // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_DATABITS; // go to next setting
				}
			}
			if (BUSVOODOO_RS232_SETTING_DATABITS==busvoodoo_rs232_setting) { // if next setting
				snprintf(busvoodoo_global_string, LENGTH(busvoodoo_global_string), "data bits (8-9) [%u]", busvoodoo_rs232_databits); // prepare next setting
				*prefix = busvoodoo_global_string; // display next setting
			}
			break;
		case BUSVOODOO_RS232_SETTING_DATABITS:
			if (NULL==line || 0==strlen(line)) { // use default setting
				busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_PARITY; // go to next setting
			} else { // setting provided
				uint8_t databits = atoi(line); // parse setting
				if (8==databits || 9==databits) { // check setting
					busvoodoo_rs232_databits = databits; // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_PARITY; // go to next setting
				}
			}
			if (BUSVOODOO_RS232_SETTING_PARITY==busvoodoo_rs232_setting) { // if next setting
				printf("1) none\n");
				printf("2) even\n");
				printf("3) odd\n");
				snprintf(busvoodoo_global_string, LENGTH(busvoodoo_global_string), "parity (1,2,3) [%c]", USART_PARITY_NONE==busvoodoo_rs232_parity ? '1' : (USART_PARITY_EVEN==busvoodoo_rs232_parity ? '2' : '3')); // prepare next setting
				*prefix = busvoodoo_global_string; // display next setting
			}
			break;
		case BUSVOODOO_RS232_SETTING_PARITY:
			if (NULL==line || 0==strlen(line)) { // use default setting
				busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_STOPBITS; // go to next setting
			} else if (1==strlen(line)) { // setting provided
				if ('1'==line[0]) { // no parity
					busvoodoo_rs232_parity = USART_PARITY_NONE;
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_STOPBITS; // go to next setting
				} else if ('2'==line[0]) { // even parity
					busvoodoo_rs232_parity = USART_PARITY_EVEN;
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_STOPBITS; // go to next setting
				} else if ('3'==line[0]) { // odd parity
					busvoodoo_rs232_parity = USART_PARITY_ODD;
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_STOPBITS; // go to next setting
				}
			}
			if (BUSVOODOO_RS232_SETTING_STOPBITS==busvoodoo_rs232_setting) { // if next setting
				printf("1) 0.5\n");
				printf("2) 1\n");
				printf("3) 1.5\n");
				printf("4) 2\n");
				snprintf(busvoodoo_global_string, LENGTH(busvoodoo_global_string), "stop bits (1,2,3,4) [%s]", USART_STOPBITS_0_5==busvoodoo_rs232_stopbits ? "0.5" : (USART_STOPBITS_1==busvoodoo_rs232_stopbits ? "1" : (USART_STOPBITS_1_5==busvoodoo_rs232_stopbits ? "1.5" : "2.0"))); // prepare next setting
				*prefix = busvoodoo_global_string; // display next setting
			}
			break;
		case BUSVOODOO_RS232_SETTING_STOPBITS:
			if (NULL==line || 0==strlen(line)) { // use default setting
				busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_HWFLOWCTL; // go to next setting
			} else if (1==strlen(line)) { // setting provided
				if ('1'==line[0]) { // 0.5 stop bits
					busvoodoo_rs232_stopbits = USART_STOPBITS_0_5; // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_HWFLOWCTL; // go to next setting
				} else if ('2'==line[0]) { // 1 stop bits
					busvoodoo_rs232_stopbits = USART_STOPBITS_1; // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_HWFLOWCTL; // go to next setting
				} else if ('3'==line[0]) { // 1.5 stop bits
					busvoodoo_rs232_stopbits = USART_STOPBITS_1_5; // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_HWFLOWCTL; // go to next setting
				} else if ('4'==line[0]) { // 2 stop bits
					busvoodoo_rs232_stopbits = USART_STOPBITS_2; // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_HWFLOWCTL; // go to next setting
				}
			}
			if (BUSVOODOO_RS232_SETTING_HWFLOWCTL==busvoodoo_rs232_setting) { // if next setting
				printf("1) no flow control\n");
				printf("2) RTS/CTS hardware flow control\n");
				snprintf(busvoodoo_global_string, LENGTH(busvoodoo_global_string), "flow control (1,2) [%c]", busvoodoo_rs232_hwflowctl ? '2' : '1'); // prepare next setting
				*prefix = busvoodoo_global_string; // display next setting
			}
			break;
		case BUSVOODOO_RS232_SETTING_HWFLOWCTL:
			if (NULL==line || 0==strlen(line)) { // use default setting
				busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_DONE; // go to next setting
			} else if (1==strlen(line)) { // setting provided
				if ('1'==line[0] || '2'==line[0]) { // setting provided
					busvoodoo_rs232_hwflowctl = ('2'==line[0]); // remember setting
					busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_DONE; // go to next setting
				}
			}
			if (BUSVOODOO_RS232_SETTING_DONE==busvoodoo_rs232_setting) { // we have all settings, configure RS-232
				rcc_periph_clock_enable(RCC_AFIO); // enable clock for USART alternate function
				rcc_periph_clock_enable(RCC_USART(BUSVOODOO_RS232_USART)); // enable clock for USART peripheral
				usart_set_baudrate(USART(BUSVOODOO_RS232_USART), busvoodoo_rs232_baudrate); // set baud rate
				usart_set_databits(USART(BUSVOODOO_RS232_USART), busvoodoo_rs232_databits); // set data bits
				usart_set_parity(USART(BUSVOODOO_RS232_USART), busvoodoo_rs232_parity); // set parity
				usart_set_stopbits(USART(BUSVOODOO_RS232_USART), busvoodoo_rs232_stopbits); // set stop bits
				if (busvoodoo_rs232_hwflowctl) {
					usart_set_flow_control(USART(BUSVOODOO_RS232_USART), USART_FLOWCONTROL_RTS_CTS); // set RTS/CTS flow control
				} else {
					usart_set_flow_control(USART(BUSVOODOO_RS232_USART), USART_FLOWCONTROL_NONE); // set no flow control
				}
				usart_set_mode(USART(BUSVOODOO_RS232_USART), USART_MODE_TX_RX); // full-duplex communication
				rcc_periph_clock_enable(RCC_USART_PORT(BUSVOODOO_RS232_USART)); // enable clock for USART GPIO peripheral
				gpio_set_mode(USART_TX_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_OUTPUT_10_MHZ, GPIO_CNF_OUTPUT_ALTFN_PUSHPULL, USART_TX_PIN(BUSVOODOO_RS232_USART)); // setup GPIO pin USART transmit
				gpio_set_mode(USART_RX_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, USART_RX_PIN(BUSVOODOO_RS232_USART)); // setup GPIO pin USART receive (no need to pull since the transceiver will drive it)
				if (busvoodoo_rs232_hwflowctl) { // use open drain drive mode
					gpio_set_mode(USART_RTS_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_OUTPUT_10_MHZ, GPIO_CNF_OUTPUT_ALTFN_OPENDRAIN, USART_RTS_PIN(BUSVOODOO_RS232_USART)); // setup GPIO pin USART transmit
					gpio_set_mode(USART_CTS_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, USART_CTS_PIN(BUSVOODOO_RS232_USART)); // setup GPIO pin USART receive (no need to pull since the transceiver will drive it)
				}
				gpio_clear(GPIO(BUSVOODOO_RS232_EN_PORT), GPIO(BUSVOODOO_RS232_EN_PIN)); // set low to enable RS-232 transceiver receiver
				gpio_set(GPIO(BUSVOODOO_RS232_SHDN_PORT), GPIO(BUSVOODOO_RS232_SHDN_PIN)); // set high to enable RS-232 transceiver transmitter
				usart_enable(USART(BUSVOODOO_RS232_USART)); // enable USART
				led_off(); // disable LED because there is no activity
				busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_NONE; // restart settings next time
				*prefix = "RS-232"; // display mode
				busvoodoo_oled_text_left(*prefix); // set mode title on OLED display
				char* pinout_io[10] = {"GND", "5V", "3V3", "LV", NULL, NULL, NULL, NULL, NULL, NULL}; // RS-232 mode I/O pinout
				for (uint8_t i=0; i<LENGTH(pinout_io) && i<LENGTH(busvoodoo_global_pinout_io); i++) {
					busvoodoo_global_pinout_io[i] = pinout_io[i]; // set pin names
				}
				char* pinout_rscan[5] = {"HV", "Rx", "Tx", NULL, NULL};  // RS-232 mode RS/CAN pinout
				if (busvoodoo_rs232_hwflowctl) { // hardware flow control is used
					pinout_rscan[3] = "RTS"; // update pin name
					pinout_rscan[4] = "CTS"; // update pin name
				}
				for (uint8_t i=0; i<LENGTH(pinout_rscan) && i<LENGTH(busvoodoo_global_pinout_rscan); i++) {
					busvoodoo_global_pinout_rscan[i] = pinout_rscan[i]; // set pin names
				}
				char* pinout[10] = {pinout_rscan[0], pinout_io[0], pinout_rscan[1], pinout_io[2], pinout_rscan[2], pinout_io[4], pinout_rscan[3], pinout_io[6], pinout_rscan[4], pinout_io[8]}; // pinout to display
				busvoodoo_oled_text_pinout((const char**)pinout, false); // set pinout on display
				busvoodoo_oled_update(); // update display to show text and pinout
				complete = true; // configuration is complete
			}
			break;
		default: // unknown case
			busvoodoo_rs232_setting = BUSVOODOO_RS232_SETTING_NONE; // restart settings next time
			break;
	}
	return complete;
}

/** exit RS-232 mode
 */
static void busvoodoo_rs232_exit(void)
{
	usart_disable(USART(BUSVOODOO_RS232_USART)); // disable USART
	rcc_periph_clock_disable(RCC_USART(BUSVOODOO_RS232_USART)); // disable domain clock for USART peripheral
	gpio_set_mode(USART_TX_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, USART_TX_PIN(BUSVOODOO_RS232_USART)); // set pin back to floating input
	gpio_set_mode(USART_RX_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, USART_RX_PIN(BUSVOODOO_RS232_USART)); // set pin back to floating input
	gpio_set_mode(USART_RTS_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, USART_RTS_PIN(BUSVOODOO_RS232_USART)); // set pin back to floating input
	gpio_set_mode(USART_CTS_PORT(BUSVOODOO_RS232_USART), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, USART_CTS_PIN(BUSVOODOO_RS232_USART)); // set pin back to floating input
	gpio_set(GPIO(BUSVOODOO_RS232_EN_PORT), GPIO(BUSVOODOO_RS232_EN_PIN)); // set high to disable RS-232 transceiver receiver
	gpio_clear(GPIO(BUSVOODOO_RS232_SHDN_PORT), GPIO(BUSVOODOO_RS232_SHDN_PIN)); // set low to disable RS-232 transceiver transmitter
}

/** write to RS-232
 *  @param[in] value value to write
 */
static void busvoodoo_rs232_write(uint16_t value)
{
	while ((0==(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_TXE) && !user_input_available)); // wait for transmit buffer to be empty (or user to interrupt)
	if (USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_TXE) { // we can send a character
		// remove unused bits
		if (USART_PARITY_NONE==busvoodoo_rs232_parity) { // no parity bit in frame
			if (8==busvoodoo_rs232_databits) { // 8-bit frame
				value &= 0xff;
			} else { // 9-bit frame
				value &= 0x1ff;
			}
		} else { // MSb is parity bit
			if (8==busvoodoo_rs232_databits) { // 8-bit frame
				value &= 0x7f;
			} else { // 9-bit frame
				value &= 0xff;
			}
		}
		// send data
		busvoodoo_led_red_pulse(BUSVOODOO_LED_PULSE); // pulse red LED to show transmission
		usart_send(USART(BUSVOODOO_RS232_USART), value); // transmit character
		// display data send
		printf("write: '%c'/0x", value);
		if ((USART_PARITY_NONE==busvoodoo_rs232_parity) && 9==busvoodoo_rs232_databits) { // case where the final data is 9 bits long
			printf("%03x\n", value);
		} else {
			printf("%02x\n", value);
		}
	}
}

/** read from RS-232
 */
static void busvoodoo_rs232_read(void)
{
	printf("read: ");
	while (!(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_RXNE) && !user_input_available); // wait for incoming data to be available (or user input to exit)
	if ((USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_RXNE)) { // verify if data has been received
		busvoodoo_led_blue_pulse(BUSVOODOO_LED_PULSE); // enable blue LED to show reception
		// get the errors
		bool error_noise = (0!=(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_NE)); // read noise error flag
		bool error_framing = (0!=(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_FE)); // read frame error flag
		bool error_parity = (0!=(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_PE)); // read parity error flag
		uint16_t c = usart_recv(USART(BUSVOODOO_RS232_USART)); // read received character (also clears the error flags)
		// remove unused bits
		if (USART_PARITY_NONE==busvoodoo_rs232_parity) { // no parity bit in frame
			if (8==busvoodoo_rs232_databits) { // 8-bit frame
				c &= 0xff;
			} else { // 9-bit frame
				c &= 0x1ff;
			}
		} else { // MSb is parity bit
			if (8==busvoodoo_rs232_databits) { // 8-bit frame
				c &= 0x7f;
			} else { // 9-bit frame
				c &= 0xff;
			}
		}
		// display data
		printf("'%c'/0x", c);
		if ((USART_PARITY_NONE==busvoodoo_rs232_parity) && 9==busvoodoo_rs232_databits) { // case where the final data is 9 bits long
			printf("%03x ", c);
		} else {
			printf("%02x ", c);
		}
		// display errors
		printf("(");
		if (error_noise) {
			printf("noise");
		} else if (error_framing) {
			printf("framing");
		} else if (error_parity) {
			printf("parity");
		} else {
			printf("no");
		}
		printf(" error)");
	}
	printf("\n");
}

/** perform RS-232 action
 *  @param[in] action action to perform
 *  @param[in] repetition how many times to perform the action
 *  @param[in] perform the action (true) or just check it (false)
 *  @return true if the action has been performed, false if it is malformed
 */
static bool busvoodoo_rs232_action(const char* action, uint32_t repetition, bool perform)
{
	uint32_t length = strlen(action); // remember length since it will be used a number of times
	if (NULL==action || 0==length) { // there is nothing to do
		return true;
	}

	if (1==length && 'r'==action[0]) { // read data
		if (!perform) {
			return true;
		}
		for (uint32_t i=0; i<repetition; i++) {
			busvoodoo_rs232_read(); // read from RS-232
		}
	} else if (1==length && 'u'==action[0]) { // sleep us
		if (!perform) {
			return true;
		}
		printf("wait for %u us\n", repetition);
		sleep_us(repetition); // sleep
	} else if (1==length && 'm'==action[0]) { // sleep ms
		if (!perform) {
			return true;
		}
		printf("wait for %u ms\n", repetition);
		sleep_ms(repetition); // sleep
	} else if ('0'==action[0]) { // send digit
		if (1==length) { // just send 0
			if (!perform) {
				return true;
			}
			for (uint32_t i=0; i<repetition; i++) {
				busvoodoo_rs232_write(0); // write to RS-232
			}
		} else if ('x'==action[1] || 'b'==action[1]) { // send hex/binary
			return busvoodoo_rs232_action(action+1, repetition, perform); // just retry without leading 0
		} else if (action[1]>='0' && action[1]<='9') { // send decimal
			return busvoodoo_rs232_action(action+1, repetition, perform); // just retry without leading 0
		} else { // malformed action
			return false;
		}
	} else if ('x'==action[0] && length>1) { // send hexadecimal value
		for (uint32_t i=1; i<length; i++) { // check string
			if (!((action[i]>='0' && action[i]<='9') || (action[i]>='a' && action[i]<='f') || (action[i]>='A' && action[i]<='F'))) { // check for hexadecimal character
				return false; // not an hexadecimal string
			}
		}
		if (!perform) {
			return true;
		}
		uint32_t value = strtol(&action[1], NULL, 16); // get hex value
		for (uint32_t i=0; i<repetition; i++) {
			busvoodoo_rs232_write(value); // write to SPI
		}
	} else if ('b'==action[0] && length>1) { // send binary value
		for (uint32_t i=1; i<length; i++) { // check string
			if (action[i]<'0' || action[i]>'1') { // check for binary character
				return false; // not a binary string
			}
		}
		if (!perform) {
			return true;
		}
		uint32_t value = strtol(&action[1], NULL, 2); // get binary value
		for (uint32_t i=0; i<repetition; i++) {
			busvoodoo_rs232_write(value); // write to SPI
		}
	} else if (action[0]>='1' && action[0]<='9') { // send decimal value
		for (uint32_t i=1; i<length; i++) { // check string
			if (action[i]<'0' || action[i]>'9') { // check for decimal character
				return false; // not a decimal string
			}
		}
		if (!perform) {
			return true;
		}
		uint32_t value = strtol(&action[0], NULL, 10); // get decimal value
		for (uint32_t i=0; i<repetition; i++) {
			busvoodoo_rs232_write(value); // write to SPI
		}
	} else if (length>=2 && ('"'==action[0] || '\''==action[0]) && (action[length-1]==action[0])) { // send ASCII character
		if (!perform) {
			return true;
		}
		for (uint32_t r=0; r<repetition; r++) {
			for (uint32_t i=1; i<length-1; i++) { // go through string
				busvoodoo_rs232_write(action[i]); // write to SPI
			}
		}
	} else { // malformed action
		return false;
	}
	return true; // all went well
}


// command handlers

/** command to perform actions
 *  @param[in] argument actions to perform
 */
static void busvoodoo_rs232_command_actions(void* argument)
{
	if (NULL==argument || 0==strlen(argument)) {
		printf("available actions (separated by space or ,):\n");
		printf("0\twrite decimal value\n");
		printf("0x0\twrite hexadecimal value\n");
		printf("0b0\twrite binary value\n");
		printf("\"a\"/'a'\twrite ASCII characters\n");
		printf("r\tread value\n");
		printf("u/m\twait 1 us/ms\n");
		printf(":n\trepeat action n times\n");
		return;
	}

	// copy argument since it will be modified
	char* copy = calloc(strlen(argument)+1, sizeof(char));
	if (!copy) {
		while (true);
	}
	strncpy(copy, argument, strlen(argument)+1);
	// verify and perform actions
	if (!busvoodoo_global_actions(copy, false, &busvoodoo_rs232_action)) { // verify actions
		printf("malformed action(s)\n");
	} else { // action are ok
		printf("press any key to exit\n");
		busvoodoo_global_actions(argument, true, &busvoodoo_rs232_action); // perform action
		if (user_input_available) { // user interrupted flow
			user_input_get(); // discard user input
		}
	}
	free(copy); // release memory
}

/** command to transmit a string
 *  @param[in] argument string to transmit (CR+LF when none provided)
 */
static void busvoodoo_rs232_command_transmit(void* argument)
{
	if (NULL==argument || 0==strlen(argument)) { // nothing to transmit
		argument = "\r\n"; // transmit CR+LF
	}
	printf("press any key to exit\n");
	for (uint16_t i=0; ((char*)(argument))[i] && !user_input_available; i++) {
		while ((0==(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_TXE) && !user_input_available)); // wait for transmit buffer to be empty
		if (USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_TXE) { // we can send a character
			printf("%c", ((char*)(argument))[i]); // echo character to transmit
			busvoodoo_led_red_pulse(BUSVOODOO_LED_PULSE); // pulse red LED to show transmission
			usart_send(USART(BUSVOODOO_RS232_USART), ((char*)(argument))[i]); // transmit character
		}
	}
	if (user_input_available) { // user interrupted flow
		user_input_get(); // discard user input
	}
	if (strcmp(argument, "\r\n")) {
		printf("\n");
	}
}

/** command to receive data
 *  @param[in] argument in which format to display
 */
static void busvoodoo_rs232_command_receive(void* argument)
{
	bool display_hex = false; // display in hex
	bool display_bin = false; // display in bin
	if (NULL!=argument && strlen(argument)>0) {
		if (0==strcmp(argument, "h") || 0==strcmp(argument, "hex")) { // user wants hexadecimal display
			display_hex = true; // remember to display in hexadecimal
		} else if (0==strcmp(argument, "b") || 0==strcmp(argument, "bin")) { // user wants binary display
			display_bin = true; // remember to display in binary
		} else {
			printf("malformed argument\n");
			return;
		}
	}
	printf("press any key to exit\n");
	while (!user_input_available) { // check for user input to exit
		if ((USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_RXNE)) { // verify if data has been received
			uint16_t c = usart_recv(USART(BUSVOODOO_RS232_USART)); // receive character
			busvoodoo_led_blue_pulse(BUSVOODOO_LED_PULSE); // enable blue LED to show reception
			// remove unused bits (ignore parity bit)
			if (USART_PARITY_NONE==busvoodoo_rs232_parity) { // no parity bit in frame
				if (8==busvoodoo_rs232_databits) { // 8-bit frame
					c &= 0xff;
				} else { // 9-bit frame
					c &= 0x1ff;
				}
			} else { // MSb is parity bit
				if (8==busvoodoo_rs232_databits) { // 8-bit frame
					c &= 0x7f;
				} else { // 9-bit frame
					c &= 0xff;
				}
			}
			if (display_hex) { // display data in hex
				if ((USART_PARITY_NONE==busvoodoo_rs232_parity) && 9==busvoodoo_rs232_databits) { // case where the final data is 9 bits long
					printf("%03x ", c);
				} else {
					printf("%02x ", c);
				}
			} else if (display_bin) { // display data in binary
				if (USART_PARITY_NONE==busvoodoo_rs232_parity) {
					if (8==busvoodoo_rs232_databits) { // 8-bit frame
						printf("%08b ", c);
					} else { // 9-bit frame
						printf("%09b ", c);
					}
				} else { // one bit is a parity bit
					if (8==busvoodoo_rs232_databits) { // 8-bit frame
						printf("%07b ", c);
					} else { // 9-bit frame
						printf("%08b ", c);
					}
				}
			} else { // display in ASCII
				printf("%c", c); // print received character
			}
		}
	}
	user_input_get(); // discard user input
	printf("\n"); // get to next line
}

/** command to transmit and receive data
 *  @param[in] argument no argument required
 */
static void busvoodoo_rs232_command_transceive(void* argument)
{
	(void)argument; // we won't use the argument
	printf("press 5 times escape to exit\n");
	char last_c = 0; // last user character received
	uint8_t esc_count = 0; // number of times escape has press received
	while (true) { // check for escape sequence
		if (user_input_available) { // check if user wants to transmit something
			char c = user_input_get(); // get user input
			if (0x1b==c) { // user pressed escape
				if (0x1b!=last_c) { // this is the first escape press
					esc_count = 0;
				}
				esc_count++; // increment escape count
			}
			last_c = c; // remember key press
			if (esc_count<5) { // check for escape sequence
				while ((0==(USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_TXE) && !user_input_available)); // wait for transmit buffer to be empty
				if (USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_TXE) { // we can send a character
					usart_send_blocking(USART(BUSVOODOO_RS232_USART), c); // send user character
					busvoodoo_led_red_pulse(BUSVOODOO_LED_PULSE); // enable red LED to show transmission			
				}
			} else { // user wants to exit
				break; // exit infinite loop
			}
		}
		if ((USART_SR(USART(BUSVOODOO_RS232_USART)) & USART_SR_RXNE)) { // verify if data has been received
			char c = usart_recv(USART(BUSVOODOO_RS232_USART)); // receive character
			busvoodoo_led_blue_pulse(BUSVOODOO_LED_PULSE); // enable blue LED to show reception
			printf("%c", c); // print received character
		}
	}
	printf("\n"); // get to next line
}

/** command to verify incoming transmission for error
 *  @param[in] argument argument not required
 */
static void busvoodoo_rs232_command_error(void* argument)
{
	(void)argument; // argument not used
	printf("press any key to exit\n");
	while (!user_input_available) { // wait until user interrupt
		busvoodoo_rs232_read(); // read incoming data (this also checks for errors
	}
	user_input_get(); // discard user input
}

/** RS-232 menu commands */
static const struct menu_command_t busvoodoo_rs232_commands[] = {
	{
		.shortcut = 'a',
		.name = "action",
		.command_description = "perform protocol actions",
		.argument = MENU_ARGUMENT_STRING,
		.argument_description = "[actions]",
		.command_handler = &busvoodoo_rs232_command_actions,
	},
	{
		.shortcut = 'r',
		.name = "receive",
		.command_description = "show incoming data [in hexadecimal or binary]",
		.argument = MENU_ARGUMENT_STRING,
		.argument_description = "[hex|bin]",
		.command_handler = &busvoodoo_rs232_command_receive,
	},
	{
		.shortcut = 't',
		.name = "transmit",
		.command_description = "transmit ASCII text (empty for CR+LF)",
		.argument = MENU_ARGUMENT_STRING,
		.argument_description = "[text]",
		.command_handler = &busvoodoo_rs232_command_transmit,
	},
	{
		.shortcut = 'x',
		.name = "transceive",
		.command_description = "transmit and receive data",
		.argument = MENU_ARGUMENT_NONE,
		.argument_description = NULL,
		.command_handler = &busvoodoo_rs232_command_transceive,
	},
	{
		.shortcut = 'e',
		.name = "error",
		.command_description = "verify incoming transmission for errors",
		.argument = MENU_ARGUMENT_NONE,
		.argument_description = NULL,
		.command_handler = &busvoodoo_rs232_command_error,
	},
};

const struct busvoodoo_mode_t busvoodoo_rs232_mode = {
	.name = "rs232",
	.description = "Recommended Standard 232",
	.full_only = true,
	.setup = &busvoodoo_rs232_setup,
	.commands = busvoodoo_rs232_commands,
	.commands_nb = LENGTH(busvoodoo_rs232_commands),
	.exit = &busvoodoo_rs232_exit,
};
